@@ -19,247 +19,17 @@
 
 namespace
 {
-   static const char *computeShaderSource = R"(#version 310 es
+   static const char *computeShaderSource =
+   #include "compute.glsl.h"
+   ;
 
-layout (local_size_x = 1, local_size_y = 1) in;
-layout (rgba32f, binding = 0) writeonly lowp uniform image2D imgOut;
-layout (binding = 1) uniform lowp sampler2DArray atlas;
-layout (binding = 2) uniform lowp sampler2D atlasMap;
-layout (binding = 3) uniform lowp sampler2DArray atlas_dw;
-layout (binding = 4) uniform lowp sampler2D atlasMap_dw;
-uniform lowp ivec2 glyphPixels;
-uniform lowp ivec2 sizeChars;
-uniform lowp ivec3 cursorColor;
-uniform lowp ivec4 cursorPos; // .xy: current; .zw: previous
-uniform lowp int cursorStyle;
-uniform lowp ivec4 selectRect;
-uniform lowp int selectRectMode;
-uniform highp ivec2 selectDamage;
-uniform lowp int deltaFrame;
-uniform lowp int showWraps;
-uniform lowp int hasDoubleWidth;
+   static const char *vertexShaderSource =
+   #include "vertex.glsl.h"
+   ;
 
-struct Cell
-{
-   highp uint charData;
-   highp uint fg;
-   highp uint bg;
-};
-
-layout (std430, binding = 0) buffer CharVideoMem
-{
-   Cell cells[];
-} vmem;
-
-void main ()
-{
-   ivec2 charPos = ivec2 (gl_GlobalInvocationID.xy);
-   int idx = sizeChars.x * charPos.y + charPos.x;
-   Cell cell = vmem.cells[idx];
-
-   if (deltaFrame == 1)
-   {
-      uint dirty = bitfieldExtract (cell.charData, 23, 1);
-      if (dirty == 0u &&
-          charPos != cursorPos.xy && charPos != cursorPos.zw &&
-          (idx < selectDamage.x || idx >= selectDamage.y))
-         return;
-   }
-   vmem.cells[idx].charData = bitfieldInsert (cell.charData, 0u, 23, 1);
-
-   ivec2 charCode =
-      ivec2 (bitfieldExtract (cell.charData, 0, 8),  // Lowest byte
-             bitfieldExtract (cell.charData, 8, 8)); // Next-lowest byte
-
-   uint dwidth = bitfieldExtract (cell.charData, 16, 1);
-   uint dwidth_cont = bitfieldExtract (cell.charData, 17, 1);
-   if (dwidth_cont == 1u) // double-width cell continuation - drawn by left half
-      return;
-
-   if (dwidth == 1u && charPos.x < sizeChars.x - 1)
-   {
-      // check validity (dwidth_cont marker in the cell to the right)
-      if (bitfieldExtract (vmem.cells[idx + 1].charData, 17, 1) != 1u)
-         dwidth = 0u;
-   }
-
-   uint fontIdx = 0u; // 0 -> Normal; 1 -> Bold; 2 -> Italic; 3 -> BoldItalic
-   if (dwidth == 0u)
-      fontIdx = bitfieldExtract (cell.charData, 18, 2);
-   uint underline = bitfieldExtract (cell.charData, 20, 1);
-   uint inverse = bitfieldExtract (cell.charData, 21, 1);
-   uint wrap = bitfieldExtract (cell.charData, 22, 1);
-
-   ivec2 atlasPos;
-   if (dwidth == 0u)
-      atlasPos = ivec2 (vec2 (256) * texelFetch (atlasMap, charCode, 0).zw);
-   else
-      atlasPos = ivec2 (vec2 (256) * texelFetch (atlasMap_dw, charCode, 0).zw);
-
-   vec3 fgColor = vec3 (float (bitfieldExtract (cell.fg, 0, 8)),
-                        float (bitfieldExtract (cell.fg, 8, 8)),
-                        float (bitfieldExtract (cell.fg, 16, 8))) / 255.0;
-
-   vec3 bgColor = vec3 (float (bitfieldExtract (cell.bg, 0, 8)),
-                        float (bitfieldExtract (cell.bg, 8, 8)),
-                        float (bitfieldExtract (cell.bg, 16, 8))) / 255.0;
-
-   vec3 crColor = vec3 (cursorColor) / 255.0;
-
-   if (selectRectMode == 1)
-   {
-      if (charPos.y >= selectRect.y && charPos.y <= selectRect.w &&
-          charPos.x >= selectRect.x && charPos.x < selectRect.z)
-         inverse ^= 1u;
-   }
-   else if ((charPos.y > selectRect.y && charPos.y < selectRect.w) ||
-       (charPos.y == selectRect.y && charPos.x >= selectRect.x &&
-        (charPos.y < selectRect.w || charPos.x < selectRect.z)) ||
-       (charPos.y == selectRect.w && charPos.x < selectRect.z &&
-        (charPos.y > selectRect.y || charPos.x > selectRect.x)))
-      inverse ^= 1u;
-
-   if (inverse == 1u)
-   {
-      vec3 tmp = fgColor;
-      fgColor = bgColor;
-      bgColor = tmp;
-   }
-   if (crColor == bgColor)
-   {
-      crColor = vec3 (1.0) - crColor;
-   }
-   if (charPos == cursorPos.xy && cursorStyle == 1)
-   {
-      fgColor = bgColor;
-      bgColor = crColor;
-   }
-
-   ivec2 srcGlyphPixels = glyphPixels;
-   if (dwidth == 1u)
-      srcGlyphPixels = ivec2 (2, 1) * glyphPixels;
-
-   if (dwidth == 0u)
-   {  // render regular cell
-      for (int j = 0; j < glyphPixels.x; j++)
-      {
-         for (int k = 0; k < glyphPixels.y; k++)
-         {
-            ivec2 txCoords = atlasPos * srcGlyphPixels + ivec2 (j, k);
-            ivec3 txc = ivec3 (txCoords, fontIdx);
-            float lumi = texelFetch (atlas, txc, 0).r;
-            vec4 pixel = vec4 (fgColor * lumi + bgColor * (1.0 - lumi), 1.0);
-            ivec2 pxCoords = charPos * glyphPixels + ivec2 (j, k);
-            imageStore (imgOut, pxCoords, pixel);
-         }
-      }
-   }
-   else if (hasDoubleWidth == 1)
-   {  // render double-width cell
-      for (int j = 0; j < srcGlyphPixels.x; j++)
-      {
-         for (int k = 0; k < srcGlyphPixels.y; k++)
-         {
-            ivec2 txCoords = atlasPos * srcGlyphPixels + ivec2 (j, k);
-            ivec3 txc = ivec3 (txCoords, fontIdx);
-            float lumi = texelFetch (atlas_dw, txc, 0).r;
-            vec4 pixel = vec4 (fgColor * lumi + bgColor * (1.0 - lumi), 1.0);
-            ivec2 pxCoords = charPos * glyphPixels + ivec2 (j, k);
-            imageStore (imgOut, pxCoords, pixel);
-         }
-      }
-   }
-   else
-   {  // no double-width font -- draw an empty box
-      for (int j = 0; j < srcGlyphPixels.x; j++)
-      {
-         for (int k = 0; k < srcGlyphPixels.y; k++)
-         {
-            float lumi = 0.0;
-            if ((0 < j && j < srcGlyphPixels.x - 1) &&
-                (0 < k && k < srcGlyphPixels.y - 1) &&
-                (j == 1 || j == srcGlyphPixels.x - 2 ||
-                 k == 1 || k == srcGlyphPixels.y - 2))
-               lumi = 0.7;
-            vec4 pixel = vec4 (fgColor * lumi + bgColor * (1.0 - lumi), 1.0);
-            ivec2 pxCoords = charPos * glyphPixels + ivec2 (j, k);
-            imageStore (imgOut, pxCoords, pixel);
-         }
-      }
-   }
-
-   if (underline == 1u)
-   {
-      for (int j = 0; j < srcGlyphPixels.x; j++)
-      {
-         vec4 pixel = vec4 (fgColor, 1.0);
-         ivec2 pxCoords = charPos * glyphPixels +
-                          ivec2 (j, srcGlyphPixels.y - 1);
-         imageStore (imgOut, pxCoords, pixel);
-      }
-   }
-
-   if (showWraps == 1 && wrap == 1u)
-   {
-      vec4 pixel = vec4 (fgColor, 1.0);
-      for (int k = 0; k < srcGlyphPixels.y; k += 2)
-      {
-         ivec2 pxCoords = charPos * glyphPixels +
-                          ivec2 (srcGlyphPixels.x - 1, k);
-         imageStore (imgOut, pxCoords, pixel);
-      }
-   }
-
-   if (charPos == cursorPos.xy && cursorStyle == 2)
-   {
-      vec4 pixel = vec4 (crColor, 1.0);
-      for (int j = 0; j < srcGlyphPixels.x; j++)
-      {
-         ivec2 pxCoords = charPos * glyphPixels + ivec2 (j, 0);
-         imageStore (imgOut, pxCoords, pixel);
-         pxCoords += ivec2 (0, srcGlyphPixels.y - 1);
-         imageStore (imgOut, pxCoords, pixel);
-      }
-      for (int k = 1; k < srcGlyphPixels.y - 1; k++)
-      {
-         ivec2 pxCoords = charPos * glyphPixels + ivec2 (0, k);
-         imageStore (imgOut, pxCoords, pixel);
-         pxCoords += ivec2 (srcGlyphPixels.x - 1, 0);
-         imageStore (imgOut, pxCoords, pixel);
-      }
-   }
-}
-)";
-
-   static const char *vertexShaderSource = R"(#version 310 es
-
-layout (location = 0) in vec2 pos;
-layout (location = 1) in lowp vec2 vertexTexCoord;
-
-out vec2 texCoord;
-
-void main ()
-{
-   texCoord = vertexTexCoord;
-   gl_Position = vec4 (pos, 0.0, 1.0);
-}
-)";
-
-   static const char *fragmentShaderSource = R"(#version 310 es
-
-in highp vec2 texCoord;
-
-layout (rgba32f, binding = 0) readonly lowp uniform image2D imgOut;
-
-uniform highp vec2 viewPixels;
-
-layout (location = 0) out lowp vec4 outColor;
-
-void main ()
-{
-   outColor = imageLoad (imgOut, ivec2 (texCoord * viewPixels));
-}
-)";
+   static const char *fragmentShaderSource =
+   #include "fragment.glsl.h"
+   ;
 
 
    GLuint
@@ -431,7 +201,7 @@ namespace zutty
        * Setup compute program
        */
       glUseProgram (P_compute);
-      glUniform2i (compU_glyphPixels, px, py);
+      glUniform2i (compU_glyphSize, px, py);
       glUniform2i (compU_sizeChars, nCols, nRows);
       glUniform1i (compU_showWraps, opts.showWraps ? 1 : 0);
 
@@ -445,28 +215,14 @@ namespace zutty
       glCheckError ();
 
       setupAtlasTexture (reg, 0);
-
-      if (fontpk->hasBold ())
-         setupAtlasTexture (fontpk->getBold (), 1);
-      else
-         setupAtlasTexture (fontpk->getRegular (), 1);
-
-      if (fontpk->hasItalic ())
-         setupAtlasTexture (fontpk->getItalic (), 2);
-      else
-         setupAtlasTexture (fontpk->getRegular (), 2);
-
-      if (fontpk->hasBoldItalic ())
-         setupAtlasTexture (fontpk->getBoldItalic (), 3);
-      else if (fontpk->hasItalic ())
-         setupAtlasTexture (fontpk->getItalic (), 3);
-      else if (fontpk->hasBold ())
-         setupAtlasTexture (fontpk->getBold (), 3);
-      else
-         setupAtlasTexture (fontpk->getRegular (), 3);
+      setupAtlasTexture (fontpk->getBold (), 1);
+      setupAtlasTexture (fontpk->getItalic (), 2);
+      setupAtlasTexture (fontpk->getBoldItalic (), 3);
 
       setupAtlasMappingTexture (reg, GL_TEXTURE2, T_atlasMap);
 
+      glUniform2fv (compU_ulMetrics, 8, fontpk->getUlMetrics ());
+
       // Setup atlas texture for double-width characters
       if (fontpk->hasDoubleWidth ())
       {
@@ -524,9 +280,9 @@ namespace zutty
       glUniform2i (compU_sizeChars, nCols, nRows);
 
       setupTexture (GL_TEXTURE0, GL_TEXTURE_2D, T_output);
-      glTexStorage2D (GL_TEXTURE_2D, 1, GL_RGBA32F, viewWidth, viewHeight);
+      glTexStorage2D (GL_TEXTURE_2D, 1, GL_RGBA8, viewWidth, viewHeight);
       glBindImageTexture (0, T_output, 0, GL_FALSE, 0, GL_WRITE_ONLY,
-                          GL_RGBA32F);
+                          GL_RGBA8);
       glCheckError ();
 
       setupStorageBuffer <Cell> (0, B_text, nRows * nCols);
@@ -655,8 +411,9 @@ namespace zutty
       linkProgram (P_compute, "compute");
       glUseProgram (P_compute);
 
-      compU_glyphPixels = glGetUniformLocation (P_compute, "glyphPixels");
+      compU_glyphSize = glGetUniformLocation (P_compute, "glyphSize");
       compU_sizeChars = glGetUniformLocation (P_compute, "sizeChars");
+      compU_ulMetrics = glGetUniformLocation (P_compute, "ulMetrics");
       compU_cursorColor = glGetUniformLocation (P_compute, "cursorColor");
       compU_cursorPos = glGetUniformLocation (P_compute, "cursorPos");
       compU_cursorStyle = glGetUniformLocation (P_compute, "cursorStyle");
@@ -668,8 +425,9 @@ namespace zutty
       compU_hasDoubleWidth = glGetUniformLocation (P_compute, "hasDoubleWidth");
 
       logT << "compute program:"
-           << " uniform glyphPixels=" << compU_glyphPixels
+           << " uniform glyphSize=" << compU_glyphSize
            << " sizeChars=" << compU_sizeChars
+           << " ulMetrics=" << compU_ulMetrics
            << " cursorColor=" << compU_cursorColor
            << " cursorPos=" << compU_cursorPos
            << " cursorStyle=" << compU_cursorStyle
