@@ -36,6 +36,8 @@ namespace zutty
       , px (priFont.getPx ())
       , py (priFont.getPy ())
       , baseline (priFont.getBaseline ())
+      , ulTop (priFont.getUlTop ())
+      , ulThick (priFont.getUlThick ())
       , nx (priFont.getNx ())
       , ny (priFont.getNy ())
       , atlasBuf (priFont.getAtlas ())
@@ -63,6 +65,10 @@ namespace zutty
       if (c == Unicode_Replacement_Character)
          return true;
 
+      // Discard everything outside the Basic Multilingual Plane
+      if (c > std::numeric_limits<uint16_t>::max ())
+         return false;
+
       return ((dwidth && wcwidth (c) == 2) ||
               (!dwidth && wcwidth (c) < 2));
    }
@@ -81,7 +87,7 @@ namespace zutty
          throw std::runtime_error (std::string ("Failed to load font ") +
                                    filename);
 
-      /* Determine the number of glyphs to actually load, based on wcwidth ()
+      /* Determine the number of glyphs to actually load.
        * We need this number up front to compute the atlas geometry.
        */
       int num_glyphs = 0;
@@ -174,13 +180,6 @@ namespace zutty
          charcode = FT_Get_Next_Char (face, charcode, &gindex);
       }
 
-      if (loadSkipCount)
-      {
-         logI << "Skipped loading " << loadSkipCount << " code point(s) "
-              << "outside the Basic Multilingual Plane"
-              << std::endl;
-      }
-
       FT_Done_Face (face);
       FT_Done_FreeType (ft);
    }
@@ -239,6 +238,8 @@ namespace zutty
          px = facesize.width;
          py = facesize.height;
          baseline = 0;
+         ulTop = py - 1.0;
+         ulThick = 1.0;
       }
       logI << "Glyph size " << px << "x" << py << std::endl;
 
@@ -248,31 +249,57 @@ namespace zutty
       if (!overlay && face->height)
       {
          // If we are loading a fixed bitmap strike of an otherwise scaled
-         // font, we need the baseline metric.
-         baseline = round (py * (double)face->ascender / face->height);
+         // font, we need the baseline as well as the underline metrics.
+         double tpy_asc = opts.fontsize *
+            (double)face->ascender / face->units_per_EM;
+         baseline = trunc (tpy_asc);
+
+         double utop = opts.fontsize *
+            (double)face->underline_position / face->units_per_EM;
+         double uthick = opts.fontsize *
+            (double)face->underline_thickness / face->units_per_EM;
+         ulTop = baseline - utop - uthick / 2.0;
+         ulThick = uthick;
+
+         logI << "Baseline " << baseline
+              << ", underline top at " << ulTop << " thickness " << ulThick
+              << std::endl;
       }
    }
 
    void Font::loadScaled (const FT_Face& face)
    {
       logI << "Pixel size " << (int)opts.fontsize << std::endl;
-      if (FT_Set_Pixel_Sizes (face, opts.fontsize, opts.fontsize))
+      if (FT_Set_Pixel_Sizes (face, 0, opts.fontsize))
          throw std::runtime_error ("Could not set pixel sizes");
 
       double tpx = opts.fontsize *
          (double)face->max_advance_width / face->units_per_EM;
-      double tpy = tpx * face->height / face->max_advance_width + 1;
+      double tpy_asc = opts.fontsize *
+         (double)face->ascender / face->units_per_EM;
+      double tpy_desc = opts.fontsize *
+         (double)face->descender / face->units_per_EM;
       if (!overlay && !dwidth)
       {
-         px = round (tpx);
-         py = round (tpy);
+         px = trunc (tpx);
+         py = trunc (tpy_asc) + trunc (-tpy_desc);
       }
       if (!overlay)
       {
-         baseline = round (tpy * face->ascender / face->height);
+         baseline = trunc (tpy_asc);
       }
       logI << "Glyph size " << px << "x" << py << ", baseline " << baseline
            << std::endl;
+
+      double utop = opts.fontsize *
+         (double)face->underline_position / face->units_per_EM;
+      double uthick = opts.fontsize *
+         (double)face->underline_thickness / face->units_per_EM;
+      ulTop = baseline - utop - uthick / 2.0;
+      ulThick = uthick;
+
+      logI << "Underline top at " << ulTop << " thickness " << ulThick
+           << std::endl;
    }
 
    void Font::loadFace (const FT_Face& face, FT_ULong c)
@@ -288,16 +315,6 @@ namespace zutty
 
    void Font::loadFace (const FT_Face& face, FT_ULong c, const AtlasPos& apos)
    {
-      if (c > std::numeric_limits<uint16_t>::max ())
-      {
-        #ifdef DEBUG
-         logT << "Skip loading code point 0x" << std::hex << c << std::dec
-              << " outside the Basic Multilingual Plane" << std::endl;
-        #endif
-         ++loadSkipCount;
-         return;
-      }
-
       if (FT_Load_Char (face, c, FT_LOAD_RENDER))
       {
          throw std::runtime_error (
@@ -317,8 +334,8 @@ namespace zutty
 
       // raw/rasterized bitmap dimensions
       const auto& bmp = face->glyph->bitmap;
-      const int bh = std::min ({(int)bmp.rows, (int)py, py - dy + sh});
-      const int bw = std::min ({(int)bmp.width, (int)px, px - dx + sw});
+      const int bh = std::min ({(int)bmp.rows - sh, py - dy});
+      const int bw = std::min ({(int)bmp.width - sw, px - dx});
 
       const int atlas_row_offset = nx * px * py;
       const int atlas_glyph_offset = apos.y * atlas_row_offset + apos.x * px;
@@ -367,11 +384,11 @@ namespace zutty
          }
          break;
       case FT_PIXEL_MODE_GRAY:
-         for (int j = sh; j < bh; ++j)
+         for (int j = 0; j < bh; ++j)
          {
-            bmp_src_row = bmp.buffer + j * bmp.pitch + sw;
+            bmp_src_row = bmp.buffer + (j + sh) * bmp.pitch + sw;
             atl_dst_row = atlasBuf.data () + atlas_write_offset + j * nx * px;
-            for (int k = sw; k < bw; ++k)
+            for (int k = 0; k < bw; ++k)
             {
                *atl_dst_row++ = *bmp_src_row++;
             }
